#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "font.h"

#define	BUFFSIZE	1024
#define	TRUE		(-1)
#define	FALSE		0

// convert hexa-decimal character to integer
static int	hex2i(char c)
{
	if (c >= '0' && c <= '9') {
		return c - '0';
	}
	if (c >= 'A' && c <= 'F') {
		return c - 'A' + 10;
	}
	if (c >= 'a' && c <= 'f') {
		return c - 'a' + 10;
	}
	return -1;
}

static int	my_strcmp(char *str1, char *str2)
{
	return strncmp(str1, str2, strlen(str2));
}

Font *	font_init(char *filename)
{
	FILE		*fp;
	Font		*pfont = NULL;
	char		buff[BUFFSIZE];
	char		data[BUFFSIZE];
	int		dummy[2];
	int		code;
	int		data_flag = FALSE;
	int		i;
	int		j;
	uint32_t	u32;

	if ((pfont = malloc(sizeof(Font))) == NULL) {
		fprintf(stderr, "malloc() failed\n");
		exit(1);
	}
	pfont->code_start = -1;
	pfont->code_end = -1;
	chunk_init(&(pfont->chunk), sizeof(uint32_t));

	// input file handling
	if ((fp = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "fopen(%s) failed.\n", filename);
		exit(1);
	}

	// process each line of BDF file
	while (fgets(buff, BUFFSIZE, fp) != NULL) {
		if (data_flag) {
			// handle each pixel patterns
			if (my_strcmp(buff, "ENDCHAR") == 0) {
				data_flag = FALSE;
			} else {
				if (sscanf(buff, "%s", data) == 1) {
					u32 = 0;
					for (i = 0; i < pfont->pixel_width; i += 4) {
						code = hex2i(data[i / 4]);
						if (code < 0) {
							break;
						}
						for (j = 0; j < 4; j++) {
							if (i + j >= pfont->pixel_width) {
								break;
							}
							u32 <<= 1;
							if (code & (1 << (3 - j))) {
								u32 |= 1;
							}
						}
					}
					chunk_add(&pfont->chunk, &u32);
				} else {
					fprintf(stderr, "code read failed\n");
					exit(1);
				}
			}
			continue;
		}

		// read parrameters
		if (my_strcmp(buff, "FONTBOUNDINGBOX") == 0) {
			// get pixel pattern size
			if (sscanf(buff, "FONTBOUNDINGBOX %d %d %d %d", &pfont->pixel_width, &pfont->pixel_height, &dummy[0], &dummy[1]) == 4) {
				if (pfont->pixel_width > 32) {
					fprintf(stderr, "width %d too wide\n", pfont->pixel_width);
					exit(1);
				}
			} else {
				fprintf(stderr, "FONTBOUNDINGBOX params lost\n");
				exit(1);
			}
		} else if (my_strcmp(buff, "ENCODING") == 0) {
			// read character code
			if (sscanf(buff, "ENCODING %d", &u32) == 1) {
				chunk_add(&pfont->chunk, &u32);
				if (pfont->code_start == -1 || pfont->code_start > u32) {
					pfont->code_start = u32;
				}
				if (pfont->code_end == -1 || pfont->code_end < u32) {
					pfont->code_end = u32;
				}
			} else {
				fprintf(stderr, "code not found\n");
				exit(1);
			}
		} else if (my_strcmp(buff, "BITMAP") == 0) {
			// below lines are pixel patterns
			data_flag = TRUE;
		}
	}
	fclose(fp);
	u32 = 0;
	chunk_add(&pfont->chunk, &u32);

	return pfont;
}

uint32_t *	font_search(Font *p_font, uint32_t code)
{
	int		idx;
	uint32_t	*p_u32;
	if ( p_font == NULL || code < p_font->code_start || code > p_font->code_end) {
		return NULL;
	}
	p_u32 = p_font->chunk.array;
	for (idx = 0; p_u32[idx] != 0; idx += p_font->pixel_height + 1) {
		if (p_u32[idx] == code) {
			return p_u32 + idx + 1;
		}
	}
	return NULL;
}

void	font_clear(Font *p_font)
{
	if (p_font == NULL) {
		return;
	}
	free(p_font->chunk.array);
	chunk_clear(&p_font->chunk);
	free(p_font);
}
