#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iconv.h>
#include "font.h"

#define	BUFF_LEN	1024
#define	ESC		0x1b
#define	TRUE		(-1)
#define	FALSE		0
#define	MY_LANG_CODE	"UTF-8"

static char	buff[BUFF_LEN];
static Font *	p_fonts[2] = {NULL, NULL};

typedef struct s_pixels {
	int	width;
	int	height;
	uint32_t	*p_pixel;
} Pixels;

static Pixels	pixels[BUFF_LEN];

void	kanji_out(int len)
{
	int	i, j, k;
	int	out_idx = 0;
	int	kanji_mode = FALSE;
	int	max_height = 0;
	uint32_t	code;
	uint32_t	*p_u32;

	// search character data
	for (i = 0; i < len; i++) {
		if (buff[i] == '\n') {
			continue;
		}
		if (buff[i] == ESC && i <= len - 3 && buff[i + 2] == 'B') {
			switch (buff[i + 1]) {
			case '$':
				kanji_mode = TRUE;
				i += 2;
				continue;
			case '(':
				kanji_mode = FALSE;
				i += 2;
				continue;
			}
		}
		if (kanji_mode && i <= len - 2) {
			code = (uint32_t)(buff[i] * 256 + buff[i + 1]);
			i++;
		} else {
			code = (uint32_t)buff[i];
		}
		for (j = 0; j < 2; j++) {
			if ((p_u32 = font_search(p_fonts[j], code)) != NULL) {
				pixels[out_idx].width = p_fonts[j]->pixel_width;
				pixels[out_idx].height = p_fonts[j]->pixel_height;
				pixels[out_idx].p_pixel = p_u32;
				if (max_height < pixels[out_idx].height) {
					max_height = pixels[out_idx].height;
				}
				out_idx++;
			}
		}
	}

	// print out
	for (i = 0; i < max_height; i++) {
		for (j = 0; j < out_idx; j++) {
			if (pixels[j].height <= i) {
				for (k = 0; k < pixels[j].width; k++) {
					putchar(' ');
				}
			} else {
				for (k = 0; k < pixels[j].width; k++) {
					putchar( (pixels[j].p_pixel[i] & (1 << pixels[j].width - 1 - k)) ? '*' : ' ');
				}
			}
		}
		putchar('\n');
	}
}

void	my_convert(iconv_t cd, char *inbuf)
{
	size_t	inbytesleft = strlen(inbuf);
	size_t	outbytesleft = BUFF_LEN;
	size_t	resultlen;
	char	*pchar1 = inbuf;
	char	*pchar2 = buff;

	resultlen = iconv(cd, &pchar1, &inbytesleft, &pchar2, &outbytesleft);
	if ((long)resultlen < 0) {
		perror("iconv() returned error");
		exit(1);
	}
	kanji_out(BUFF_LEN - outbytesleft);
}

int	main(int argc, char *argv[])
{
	iconv_t	iconvt;
	char	inbuff[BUFF_LEN];

	iconvt = iconv_open("ISO-2022-JP", MY_LANG_CODE);
	if (iconvt < 0) {
		perror("iconv_open()");
		exit(1);
	}
	if (argc == 3) {
		p_fonts[0] = font_init(argv[1]);
		my_convert(iconvt, argv[2]);
		iconv_close(iconvt);
		return 0;
	} else if (argc == 4) {
		p_fonts[0] = font_init(argv[1]);
		p_fonts[1] = font_init(argv[2]);
		my_convert(iconvt, argv[3]);
		iconv_close(iconvt);
		return 0;
	} else if (argc > 3) {
		fprintf(stderr, "too many parameters\n");
		exit(1);
	} else {
		fprintf(stderr, "too few parameters\n");
		exit(1);
	}
	while (fgets(inbuff, BUFF_LEN, stdin) != NULL) {
		my_convert(iconvt, inbuff);
	}
	iconv_close(iconvt);
	font_clear(p_fonts[0]);
	font_clear(p_fonts[1]);
	return 0;
}
